"""
Created by: Taylor Denouden
Organization: Hakai Institute
Date: 2020-06-12
Description: Classes for processing images using Tom's bin-based glint masking technique for various types of image
    files.
"""

import re
from abc import ABCMeta, abstractmethod
from pathlib import Path
from typing import Iterable, Union

import numpy as np
from scipy.ndimage import gaussian_filter

from .abstract_masker import Masker

EPSILON = 1e-8


class BinMasker(Masker, metaclass=ABCMeta):
    """ Abstract class for all maskers that use Tom Bell's binning algorithm.
        Defines behaviour common to all Bin maskers."""

    def __init__(self, img_dir: str, out_dir: str, glint_threshold: float = 0.875, mask_buffer_sigma: int = 0) -> None:
        """Create and return a glint mask for RGB imagery.

        Parameters
        ----------
        img_dir
            The path to a directory containing images to process.
        out_dir
            Path to the directory where the image masks should be saved.
        glint_threshold
            The amount of binned "blueness" that should be glint. Domain for values is (0.0, 1.0).
        mask_buffer_sigma
            The sigma for the Gaussian kernel used to buffer the mask.
        """
        super().__init__(img_dir, out_dir)

        self.glint_threshold = glint_threshold
        self.mask_buffer_sigma = mask_buffer_sigma

    def mask_img(self, img: np.ndarray) -> np.ndarray:
        """Generates and saves a glint mask for the image at path img_path using Tom's method.

        Parameters
        ----------
        img
            The image data normalized to range [0,1].

        Returns
        -------
        np.ndarray
            The calculated mask.
        """
        # TOM's method:
        # bins = np.linspace(0., 1., num_bins, endpoint=False)
        # si = np.digitize(img, bins)
        # sis = (si.astype(np.float) - si.min(initial=0)) / ((si.max(initial=0) - si.min(initial=0)) + EPSILON)
        #
        # # Find Glint Threshold and Set those Pixels to 1
        # mask = (sis <= glint_threshold)

        # Much more efficient method.
        # With glint_threshold=0.875, it is equivalent to the above with num_bins=8 and glint_threshold=0.9
        return img < self.glint_threshold

    def postprocess_mask(self, mask: np.ndarray) -> np.ndarray:
        # Buffer the mask
        if self.mask_buffer_sigma > 0:
            mask_buffered = gaussian_filter(mask.astype(np.float), self.mask_buffer_sigma)
            mask = mask_buffered >= 0.99

        # Convert to format required by Metashape
        return mask.astype(np.uint8) * 255


class RGBBinMasker(BinMasker):
    """Tom Bell's method masker for RGB imagery."""

    def preprocess_img(self, img: np.ndarray) -> np.ndarray:
        """Selects the blue channel and normalize to [0, 1]."""
        return self.normalize_img(img[:, :, 2], bit_depth=8)

    def get_mask_save_paths(self, in_path: str) -> Iterable[str]:
        """Implement get_out_paths as required by AbstractBinMasker."""
        return [str(Path(self.out_dir).joinpath(f"{Path(in_path).stem}_mask.png"))]


class _P4MSBinMasker(BinMasker, metaclass=ABCMeta):
    """ Tom Bell method masker for DJI multi-spectral imagery."""

    @property
    @abstractmethod
    def filename_matcher(self):
        raise NotImplemented

    def is_dji_red_edge(self, filename: str) -> bool:
        """Determine if the filename belongs to a DJI multispectral red edge image."""
        return self.filename_matcher.match(str(filename)) is not None

    @property
    def img_paths(self) -> Iterable[str]:
        """Generates an iterator with the files to process.

        For DJI Multispectral masking, this is a list of paths that correspond to the RedEdge band of the files output
        from a DJI multispectral camera.
        """
        return filter(self.is_dji_red_edge, super().img_paths)

    def preprocess_img(self, img: np.ndarray) -> np.ndarray:
        """Normalize 16-bit images to range [0,1]."""
        return self.normalize_img(img, bit_depth=16)

    def get_mask_save_paths(self, in_path: str) -> Iterable[str]:
        """Generates a list of output mask paths for each input image file path.

        For DJI multispectral, we save a mask file for each of the bands even though only the red edge band was used to
        generate the masks.
        """
        in_path_root = Path(in_path).stem[:-1]
        out_dir = Path(self.out_dir)
        return [str(out_dir.joinpath(f"{in_path_root}{i}_mask.png")) for i in range(1, 6)]


class P4MSRedEdgeBinMasker(_P4MSBinMasker):
    """ Tom Bell method masker for DJI multi-spectral imagery."""

    @property
    def filename_matcher(self):
        return re.compile("(.*[\\\\/])?DJI_[0-9]{3}4.TIF", flags=re.IGNORECASE)


class P4MSBlueBinMasker(_P4MSBinMasker):
    """ Tom Bell method masker for DJI multi-spectral imagery."""

    @property
    def filename_matcher(self):
        return re.compile("(.*[\\\\/])?DJI_[0-9]{3}1.TIF", flags=re.IGNORECASE)


class _MicasenseRedEdgeBinMasker(BinMasker, metaclass=ABCMeta):
    """Tom Bell method masker for Micasense RedEdge Camera imagery."""

    @property
    @abstractmethod
    def filename_matcher(self):
        raise NotImplemented

    def is_micasense_red_edge(self, filename: Union[Path, str]) -> bool:
        """Determine if the filename belongs to a Micasense red edge image."""
        return self.filename_matcher.match(str(filename)) is not None

    @property
    def img_paths(self) -> Iterable[str]:
        """Generates the list of files to process.

        For Micasense Red Edge masking, this should be a list of paths that correspond to the RedEdge band of the files
        output from a Micasense camera.
        """
        # Filter out all but the red edge band files
        return filter(self.is_micasense_red_edge, super().img_paths)

    def preprocess_img(self, img: np.ndarray) -> np.ndarray:
        """Normalize 16-bit imagery to have values in the range [0,1]."""
        return self.normalize_img(img, bit_depth=16)

    def get_mask_save_paths(self, in_path: str) -> Iterable[str]:
        """Generates a list of output mask paths for each input image file path.

        For Micasense, we wants a mask file for each of the bands even though only the red edge band was used to
        generate the masks."""
        in_path_root = Path(in_path).stem[:-1]
        out_dir = Path(self.out_dir)
        return [str(out_dir.joinpath(f"{in_path_root}{i}_mask.png")) for i in range(1, 6)]


class MicasenseRedEdgeBinMasker(_MicasenseRedEdgeBinMasker):
    """Tom Bell method masker for Micasense RedEdge Camera imagery."""

    @property
    def filename_matcher(self):
        return re.compile("(.*[\\\\/])?IMG_[0-9]{4}_5.tif", flags=re.IGNORECASE)


class MicasenseBlueBinMasker(_MicasenseRedEdgeBinMasker):
    """Tom Bell method masker for Micasense RedEdge Camera imagery."""

    @property
    def filename_matcher(self):
        return re.compile("(.*[\\\\/])?IMG_[0-9]{4}_1.tif", flags=re.IGNORECASE)



"""
Created by: Taylor Denouden
Organization: Hakai Institute
Date: 2020-06-12
Description: Classes for generating glint masks using the specular reflection estimation technique for various
    types of image files.
"""
import math
from abc import ABCMeta
from pathlib import Path
from typing import Iterable

import numpy as np
from scipy.ndimage.morphology import binary_closing, binary_opening

from .abstract_masker import Masker

# For numerical stability in division ops
EPSILON = 1e-8


class SpecularMasker(Masker, metaclass=ABCMeta):
    """ Specular masker method for RGB imagery."""

    def __init__(self, img_dir: str, out_dir: str, percent_diffuse: float = 0.95, mask_thresh: float = 0.99,
                 opening: int = 15, closing: int = 15) -> None:
        """Create and return a glint mask for RGB imagery.

        Parameters
        ----------
        img_dir
            The path to a directory containing images to process.
        out_dir
            Path to the directory where the image masks should be saved.
        percent_diffuse
            An estimate of the percentage of pixels in an image that show pure diffuse reflectance, and thus no specular
            reflectance (glint).
        mask_thresh
            The threshold on the specular reflectance estimate image to convert into a mask.
            e.g. if more than 50% specular reflectance is unacceptable, use 0.5.
        opening
            The number of morphological opening iterations on the produced mask.
            Useful for closing small holes in the mask.
        closing
            The number of morphological closing iterations on the produced mask.
            Useful for removing small bits of mask.
        """
        super().__init__(img_dir, out_dir)

        self.percent_diffuse = percent_diffuse
        self.mask_thresh = mask_thresh
        self.opening = opening
        self.closing = closing

    def mask_img(self, img: np.ndarray) -> np.ndarray:
        """Create and return a glint mask for RGB imagery.

        Parameters
        ----------
        img: np.ndarray shape=(H,W,3)
            Path to a 3-channel RGB numpy image normalized to values in [0,1].

        Returns
        -------
        numpy.ndarray, shape=(H,W)
            Numpy array of glint mask for img at input_path.
        """
        return self.estimate_specular_reflection_component(img, self.percent_diffuse) >= self.mask_thresh

    def postprocess_mask(self, mask: np.ndarray) -> np.ndarray:
        # Fill in small holes in the mask
        if self.opening > 0:
            mask = binary_opening(mask, iterations=self.opening).astype(np.uint8)

        # Remove small bits of mask
        if self.closing > 0:
            mask = binary_closing(mask, iterations=self.closing).astype(np.uint8)

        # Save the mask
        return mask * 255

    @staticmethod
    def estimate_specular_reflection_component(img: np.ndarray, percent_diffuse: float) -> np.ndarray:
        """Estimate the specular reflection component of pixels in an image.

            Based on method from:
                Wang, S., Yu, C., Sun, Y. et al. Specular reflection removal
                of ocean surface remote sensing images from UAVs. Multimedia Tools
                Appl 77, 11363–11379 (2018). https://doi.org/10.1007/s11042-017-5551-7

            Parameters
            ----------
            img: numpy.ndarray, shape=(H,W,C)
                A numpy ndarray of an RGB image.
            percent_diffuse
                An estimate of the % of pixels that show purely diffuse reflection.

            Returns
            -------
            numpy.ndarray, shape=(H,W)
                An 1D image where values are an estimate of the component of specular reflectance.
            """
        # Calculate the pixel-wise max intensity and intensity range over RGB channels
        i_max = np.amax(img, axis=2)
        i_min = np.amin(img, axis=2)
        i_range = i_max - i_min

        # Calculate intensity ratio
        q = np.divide(i_max, i_range + EPSILON)

        # Select diffuse only pixels using the PERCENTILE_THRESH
        # i.e. A percentage of PERCENTILE_THRESH pixels are supposed to have no
        #     specular reflection
        num_thresh = math.ceil(percent_diffuse * q.size)

        # Get intensity ratio for a pixel dividing the image into diffuse and specular sections
        q_x_hat = np.partition(q.ravel(), num_thresh)[num_thresh]

        # Estimate the spectral component of each pixel
        spec_ref_est = np.clip(i_max - (q_x_hat * i_range), 0, None)

        return spec_ref_est


class RGBSpecularMasker(SpecularMasker):
    """ Specular masker method for RGB imagery."""

    def preprocess_img(self, img: np.ndarray) -> np.ndarray:
        """Normalizes 8-bit pixel values and select only the RGB channels."""
        return self.normalize_img(img[:, :, :3], bit_depth=8)

    def get_mask_save_paths(self, in_path: str) -> Iterable[str]:
        """Get the out path for where to save the mask corresponding to image at in_path."""
        return [str(Path(self.out_dir).joinpath(f"{Path(in_path).stem}_mask.png"))]



"""
Created by: Taylor Denouden
Organization: Hakai Institute
Date: 2020-09-16
Description: 
"""

import re
from abc import ABCMeta, abstractmethod
from pathlib import Path
from typing import Iterable, Pattern, Sequence

import numpy as np
from PIL import Image
from scipy.ndimage import gaussian_filter

from core.abstract_masker import Masker

EPSILON = 1e-8


class ThresholdMasker(Masker, metaclass=ABCMeta):
    """ Abstract class for all maskers that use threshold masking algorithm.
        Defines behaviour common to all Bin maskers."""

    def __init__(self, img_dir: str, out_dir: str,
                 glint_thresholds: Iterable[float], mask_buffer_sigma: int = 0) -> None:
        """Create and return a glint mask for RGB imagery.

        Parameters
        ----------
        img_dir
            The path to a directory containing images to process.
        out_dir
            Path to the directory where the image masks should be saved.
        glint_thresholds
            Threshold for each band where values with pixels above this are considered glint. Masks are generated for
            each band and then combined using a union operation. Values must be in domain [0., 1.].
        mask_buffer_sigma
            The sigma for the Gaussian kernel used to buffer the mask.
        """
        super().__init__(img_dir, out_dir)

        self.glint_thresholds = glint_thresholds
        self.mask_buffer_sigma = mask_buffer_sigma

    def mask_img(self, img: np.ndarray) -> np.ndarray:
        """Generates and saves a glint mask for the image at path img_path using Tom's method.

        Parameters
        ----------
        img
            The image data normalized to range [0,1]. Channel order is H, W, C

        Returns
        -------
        np.ndarray
            The calculated mask.
        """
        return np.any(img < self.glint_thresholds, axis=2)

    def postprocess_mask(self, mask: np.ndarray) -> np.ndarray:
        """Buffer the mask and convert to format required by Agisoft."""
        # Buffer the mask
        if self.mask_buffer_sigma > 0:
            mask_buffered = gaussian_filter(mask.astype(np.float), self.mask_buffer_sigma)
            mask = mask_buffered >= 0.99

        # Convert to format required by Metashape
        return mask.astype(np.uint8) * 255


class GenericThresholdMasker(ThresholdMasker):
    """Threshold masker for RGB and ACO imagery."""

    def preprocess_img(self, img: np.ndarray) -> np.ndarray:
        """Normalize to [0, 1] and switch band order to BGR."""
        # e.g. [0,255] -> [0,1]
        return self.normalize_img(img, bit_depth=8)

    def get_mask_save_paths(self, in_path: str) -> Iterable[str]:
        """Implement get_out_paths as required by AbstractBinMasker."""
        return [str(Path(self.out_dir).joinpath(f"{Path(in_path).stem}_mask.png"))]


class _MultiFileThresholdMasker(ThresholdMasker, metaclass=ABCMeta):
    """Threshold masker for imagery split over multiple files."""

    @property
    @abstractmethod
    def _file_matcher(self) -> Pattern:
        raise NotImplemented

    def is_first_band_file(self, filename: str) -> bool:
        """Determine if the filename belongs to a P4MS blue band image."""
        return self._file_matcher.match(str(filename)) is not None

    @property
    def img_paths(self) -> Iterable[str]:
        return filter(self.is_first_band_file, super().img_paths)

    def preprocess_img(self, img: np.ndarray) -> np.ndarray:
        """Normalize 16-bit images to range [0,1]."""
        return self.normalize_img(img, bit_depth=16)

    def read_img(self, img_path: str) -> np.ndarray:
        """Overrides baseclass image reader to read multiple files into a single numpy array.

        Parameters
        ----------
        img_path
            The path to the first image of 5.

        Returns
        -------
        np.ndarray
            The image data in a np.ndarray with dtype=float32.
        """
        img_paths = self.get_grouped_img_paths(img_path)
        imgs = [np.asarray(Image.open(p)) for p in img_paths]
        result = np.stack(imgs, axis=2).astype(np.float)
        return result

    @staticmethod
    @abstractmethod
    def get_grouped_img_paths(first_band_path: str) -> Sequence[str]:
        raise NotImplemented


class P4MSThresholdMasker(_MultiFileThresholdMasker):
    """Threshold masker for DJI P4MS imagery."""

    @property
    def _file_matcher(self) -> Pattern:
        return re.compile("(.*[\\\\/])?DJI_[0-9]{3}1.TIF", flags=re.IGNORECASE)

    @staticmethod
    def get_grouped_img_paths(first_band_path: str) -> Sequence[str]:
        """Returns a tuple with the files to process.

        e.g. DJI P4MS, this is a list of paths that correspond to the all files output
        from a DJI multispectral camera for a single capture event.
        """
        in_path_root = Path(first_band_path).stem[:-1]
        return [str(Path(first_band_path).with_name(f"{in_path_root}{i}.TIF")) for i in range(1, 6)]

    def get_mask_save_paths(self, in_path: str) -> Iterable[str]:
        """Generates a list of output mask paths for each input image file path.

        For DJI multispectral, we save a mask file for each of the bands even though only the red edge band was used to
        generate the masks.
        """
        in_path_root = Path(in_path).stem[:-1]
        out_dir = Path(self.out_dir)
        return (str(out_dir.joinpath(f"{in_path_root}{i}_mask.png")) for i in range(1, 6))


class MicasenseRedEdgeThresholdMasker(_MultiFileThresholdMasker):
    """Threshold masker for MicasenseRedEdge imagery."""

    @property
    def _file_matcher(self) -> Pattern:
        return re.compile("(.*[\\\\/])?IMG_[0-9]{4}_1.tif", flags=re.IGNORECASE)

    @staticmethod
    def get_grouped_img_paths(first_band_path: str) -> Sequence[str]:
        """Returns a tuple with the files to process.

        e.g. DJI P4MS, this is a list of paths that correspond to the all files output
        from a DJI multispectral camera for a single capture event.
        """
        in_path_root = Path(first_band_path).stem[:-1]
        return [str(Path(first_band_path).with_name(f"{in_path_root}{i}.tif")) for i in range(1, 6)]

    def get_mask_save_paths(self, in_path: str) -> Iterable[str]:
        """Generates a list of output mask paths for each input image file path.

        For Micasense, we wants a mask file for each of the bands even though only the red edge band was used to
        generate the masks."""
        in_path_root = Path(in_path).stem[:-1]
        out_dir = Path(self.out_dir)
        return (str(out_dir.joinpath(f"{in_path_root}{i}_mask.png")) for i in range(1, 6))


if __name__ == '__main__':
    masker = MicasenseRedEdgeThresholdMasker(
        '/media/taylor/Samsung_T5/Datasets/ExampleImages/MicasenseRededge',
        '/tmp',
        glint_thresholds=[0.875, 1., 1., 1., 1.]
    )
    masker(max_workers=1, callback=print, err_callback=print)

    # masker = P4MSThresholdMasker(
    #     '/media/taylor/Samsung_T5/Datasets/ExampleImages/P4MS',
    #     '/tmp',
    #     glint_thresholds=[0.875, 1., 1., 1., 1.]
    # )
    # masker(max_workers=1, callback=print, err_callback=print)



"""
Created by: Taylor Denouden
Organization: Hakai Institute
Date: 2020-06-12
Description: Abstract Base Class for GlintMaskGenerators. Implements a number of functions that allows for
  multi-threaded processing of images and a simple interface for which the GUI and CLIs can be easily hook into.
"""

import concurrent.futures
import os
from abc import ABC, abstractmethod
from concurrent.futures.thread import ThreadPoolExecutor
from pathlib import Path
from typing import Callable, Iterable, Optional

import numpy as np
from PIL import Image

Image.MAX_IMAGE_PIXELS = None


class Masker(ABC):
    """Abstract class for image mask generators."""

    def __init__(self, img_dir: str, out_dir: str):
        """Constructor for AbstractBaseMasker.

        Parameters
        ----------
        img_dir
            The path to the directory where the images to process are located.
        out_dir
            The path to the directory where the produced masks should be saved.
        """
        super().__init__()

        self.img_dir = img_dir
        self.out_dir = out_dir

    def __call__(self, max_workers: int,
                 callback: Optional[Callable[[str], None]] = None,
                 err_callback: Optional[Callable[[str, Exception], None]] = None
                 ) -> None:
        """Allows calling self.process by calling the object as a function.

        Parameters
        ----------
        max_workers
            The number of threads to use during processing. Useful for limiting memory consumption.
        callback
            Optional callback that receives the img_path as an arg after it is successfully processed.
        err_callback
            Optional callback that receives the img_path and an Exception as args after a processing failure.
        """
        return self.process(max_workers, callback, err_callback)

    def __len__(self) -> int:
        """Get and return the number of files to process.

        Returns
        -------
        int
            The number of files that need to be processed.
        """
        return sum(1 for _ in self.img_paths)

    @property
    def img_paths(self) -> Iterable[str]:
        """Property that lists the file paths to process in self.img_dir.

        Returns an iterator that lists the files to process. Subclasses may want to override this to return specific
        image types or filter the results. By default, will list all images in self.img_dir if the file extension is in
        the extensions list.

        Returns
        -------
        Iterable[str]
            The list of files to be used for generating the masks.
        """
        extensions = (".png", ".jpg", ".jpeg", ".tif", ".tiff")

        paths = Path(self.img_dir).glob("**/*")
        paths = filter(lambda p: p.is_file() and p.suffix.lower() in extensions, paths)

        return map(lambda p: str(p), paths)

    # noinspection PyMethodMayBeStatic
    def preprocess_img(self, img: np.ndarray) -> np.ndarray:
        """Optional image pre-processing hook.

        Subclasses may implement this method to do preprocessing step on image before passing it to the actual masking
        function. Common use-cases are selecting appropriate bands and value normalization.

        Parameters
        ----------
        img
            The image data as a numpy array with dtype=float32.

        Returns
        -------
        np.ndarray
            The modified image which is next passed to the the masking function.

        """
        return img

    # noinspection PyMethodMayBeStatic
    def postprocess_mask(self, mask: np.ndarray) -> np.ndarray:
        """Optional mask post-processing hook.

        Subclasses may implement this method to do preprocessing step on the produced mask before it saved to the mask
        out paths.

        Parameters
        ----------
        mask
            The mask created by the glint masking algorithm.

        Returns
        -------
        np.ndarray
            The modified mask which is then saved to the mask out paths.
        """
        return mask

    @abstractmethod
    def mask_img(self, img: np.ndarray) -> np.ndarray:
        """Generate a mask for img.

        All subclasses must implement this method. Takes a numpy ndarray that has been normalized and generates a
        glint mask.

        Parameters
        ----------
        img
            The preprocessed image in numpy array format.

        Returns
        -------
        np.ndarray, shape=(H,W)
            The calculated glint mask.
        """
        raise NotImplemented

    @abstractmethod
    def get_mask_save_paths(self, in_path: str) -> Iterable[str]:
        """Get the out paths where the mask corresponding to the image at in_path should be saved.

        Parameters
        ----------
        in_path
            The image path for which a mask is generated.
            Used to generate an appropriate out path for the mask.

        Returns
        -------
        Iterable[str]
            A list of paths where the mask for the image at location in_path should be saved.
        """
        raise NotImplemented

    def process(self, max_workers: int = os.cpu_count() * 5,
                callback: Optional[Callable[[str], None]] = None,
                err_callback: Optional[Callable[[str, Exception], None]] = None
                ) -> None:
        """Compute all glint masks.

        Computes masks for all images in self.img_paths using the process_func and save to the mask_out_paths.

        Parameters
        ----------
        max_workers
            The maximum number of image processing workers.
            Useful for limiting memory usage.
        callback
            Optional callback function passed the name of each input and output mask files after processing it.
            Will receive img_path: str as arg.
        err_callback
            Optional callback function passed exception object on processing failure.
            Will receive img_path: str, and the Exception as args.
        """
        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            future_to_path = {executor.submit(self.process_one_file, path): path for path in self.img_paths}
            for future in concurrent.futures.as_completed(future_to_path):
                path = future_to_path[future]
                try:
                    _ = future.result()
                    if callback is not None:
                        callback(path)

                except Exception as exc:
                    if err_callback is not None:
                        err_callback(path, exc)
                    executor.shutdown(wait=False)
                    return

    def process_one_file(self, img_path: str) -> None:
        """Generates and saves a glint mask for the image located at img_path.

        Saves the generated mask to all path locations returned by self.get_mask_save_paths(img_path).

        Parameters
        ----------
        img_path
            The path to the image to generate a glint mask for.
        """
        img = self.read_img(img_path)
        img = self.preprocess_img(img)

        mask = self.mask_img(img)
        mask = self.postprocess_mask(mask)

        for path in self.get_mask_save_paths(img_path):
            self.save_mask(mask, path)

    @staticmethod
    def save_mask(mask: np.ndarray, out_path: str) -> None:
        """Utility function to save a mask to the location out_path.

        Parameters
        ----------
        mask
            2D image mask to save into the image format specified in the out_path.
        out_path
            The path where the file should be saved, including img extension.
        """
        mask_img = Image.fromarray(mask, mode='L')
        mask_img.save(str(out_path))

    @staticmethod
    def read_img(img_path: str) -> np.ndarray:
        """Utility function to read img at img_path into a numpy array of type float.

        Parameters
        ----------
        img_path
            The path to the image to open.

        Returns
        -------
        np.ndarray
            The image data in a np.ndarray with dtype=float32.
        """
        img = Image.open(img_path)
        return np.array(img).astype(np.float)

    @staticmethod
    def normalize_img(img: np.ndarray, bit_depth: int) -> np.ndarray:
        """Utility function to normalize the values of an image with arbitrary bit_depth to range [0, 1].

        Parameters
        ----------
        img
            The image data
        bit_depth
            The bit depth of the image.
            e.g. 8 for 8-bit, 16 for 16-bit, 32 for 32-bit.

        Returns
        -------
        np.ndarray
            The image normalized to range [0,1].
        """
        return img / ((1 << bit_depth) - 1)



"""Core functions for the GUI and CLI glint mask generators."""

__all__ = [
    'BinMasker',
    'RGBBinMasker',
    'P4MSBlueBinMasker',
    'MicasenseBlueBinMasker',
    'P4MSRedEdgeBinMasker',
    'MicasenseRedEdgeBinMasker',
    'SpecularMasker',
    'RGBSpecularMasker',
    'GenericThresholdMasker',
    'P4MSThresholdMasker',
    'MicasenseRedEdgeThresholdMasker',
    'Masker'
]

from .abstract_masker import *
from .bin_maskers import *
from .specular_maskers import *
from .threshold_maskers import *



"""
Created by: Taylor Denouden
Organization: Hakai Institute
Date: 2020-06-12
"""

from abc import ABC
from pathlib import Path

import numpy as np
from PIL import Image

from core.abstract_masker import Masker
from core.specular_maskers import RGBSpecularMasker

IMG_CONTENT = Image.fromarray(np.ones((32, 32, 3)).astype(np.uint8) * 255)


def test_class_inheritance():
    """Test that the classes have the correct superclasses."""
    assert issubclass(Masker, ABC)
    assert issubclass(RGBSpecularMasker, Masker)


def test_rgb_specular_masker(tmp_path):
    """Test that the RGB specular masker finds the correct files."""
    # Create dummy files
    all_paths = ["1.jpg", "2.JPG", "3.tif", "4.tiff", "5.TIF", "6.TIFF", "7.png", "8.PNG", "9.jpeg", "10.JPEG",
                 "IMG_1234_1.tif", "IMG_1234_2.tif", "IMG_1234_3.tif", "IMG_1234_4.tif", "IMG_1234_5.tif",
                 "IMG_4321_1.TIF", "IMG_4321_2.TIF", "IMG_4321_3.TIF", "IMG_4321_4.TIF", "IMG_4321_5.TIF",
                 "DJI_1021.TIF", "DJI_1022.TIF", "DJI_1023.TIF", "DJI_1024.TIF", "DJI_1025.TIF",
                 "DJI_2021.tif", "DJI_2022.tif", "DJI_2023.tif", "DJI_2024.tif", "DJI_2025.tif"]
    all_paths = sorted([str(tmp_path.joinpath(n)) for n in all_paths])
    for name in all_paths:
        IMG_CONTENT.save(name)

    # Test the class methods
    masker = RGBSpecularMasker(tmp_path, Path(tmp_path).joinpath("masks"))

    valid_paths = all_paths
    valid_paths = sorted([str(tmp_path.joinpath(n)) for n in valid_paths])
    assert len(masker) == len(valid_paths)
    assert all(np.array(sorted(masker.img_paths)) == np.array(sorted(valid_paths)))
